"""Metadata module."""

__title__: str = "biar"
__description__: str = "batteries-included async requests tool for python."
__version__: str = "0.5.1"
__url__: str = "https://github.com/rafaelleinio/biar"
__author__: str = "Rafael Leinio"
__license__: str = "MIT"
